/***************************************************************************
 *            Grille.cpp
 *
 *  2005  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Grille.hpp"
using namespace std;

Grille::Grille():m_name("grille"), m_active(false), m_speed(1), m_homothety(false), m_origin(ORIGIN_UP_LEFT), m_loopState(LOOP_EMPTY), m_loopTransparency(0.0), m_blend(false), m_loopPos(-1), m_audioActive(false), m_hasGlCoords(false)
{
	//default values
	float maxRows = MAX_ROWS;
	float maxCols = MAX_COLS;
	m_params[PARAM_COLS].setElementRange(0, 1, maxRows);
	m_params[PARAM_COLS].setInitValue(1.0/30.0);
	m_params[PARAM_ROWS].setElementRange(0, 1, maxCols);
	m_params[PARAM_ROWS].setInitValue(1.0/30.0);
	m_params[PARAM_IMAGE].setElementRange(0,0,1);
	m_params[PARAM_IMAGE].setValue(0);
	m_params[PARAM_SIZE].setElementRange(0,0,1);
	m_params[PARAM_SIZE].setInitValue(0.8);
	m_params[PARAM_TRANS].setElementRange(0,0,1);
	m_params[PARAM_TRANS].setInitValue(0);
	m_params[PARAM_ROTATION].setElementRange(0,0,360);
	m_params[PARAM_ROTATION].setInitValue(0);
	std::string rangeMin="255;255;255";
	std::string rangeMax="0;0;255";
	m_params[PARAM_COLOR].setElementColorRange(0, rangeMin, rangeMax);
	m_params[PARAM_COLOR].setInitValue(0);
	/*
	m_params[PARAM_DISX].setElementRange(0,0,100);
	m_params[PARAM_DISX].setInitValue(0);
	m_params[PARAM_DISY].setElementRange(0,0,100);
	m_params[PARAM_DISY].setInitValue(0);
	*/
}


Grille::~Grille()
{
	
}

void Grille::copy(Grille& copiedGrille)
{
	//copy grille
	m_name=copiedGrille.getName()+"-copy";	
	m_homothety=copiedGrille.getHomothety();	
	m_origin=copiedGrille.getOrigin();	
	m_blend=copiedGrille.m_blend;	

	//copy parameters
	for(unsigned int i=0;i<PARAMS_NB;++i) {
		m_params[VJ_PARAM(i)].copy(copiedGrille.getParameter(VJ_PARAM(i)));
	}

	m_loopState=copiedGrille.m_loopState;
	m_loopStartTime=copiedGrille.m_loopStartTime;
	m_loopDuration=copiedGrille.m_loopDuration;
}

void Grille::grabInitValues()
{
	for(unsigned int i=0;i<PARAMS_NB;++i) {
		m_params[VJ_PARAM(i)].grabInitValues();
	}
}


void Grille::setElementAnim(const int& element, Anim *anim)
{
	//set the anim
	m_params[PARAM_IMAGE].setElementAnim(element,anim);
}

const float& Grille::getParameterValueAt(const VJ_PARAM& p, const int& pos)
{
	return m_params[p].getValueAt(pos);
}

const GLuint& Grille::getImageAt(const int& pos)
{
	return m_params[PARAM_IMAGE].getImageAt(pos);
}

void Grille::setElementColorRange(const int& el, std::string& min, std::string& max)
{
	m_params[PARAM_COLOR].setElementColorRange(el, min, max);
}


bool Grille::playLoop(const long& time)
{
	bool changed=false;
	if(m_loopState==LOOP_PLAYING || m_loopState==LOOP_OVERDUBBING) {
		for(unsigned int i=0;i<PARAMS_NB;++i) {
			if(m_params[VJ_PARAM(i)].playLoop(time)) {
				changed=true;
			}
		}
		m_loopPos = (float(time-m_loopStartTime)*m_speed)/float(m_loopDuration);
		if(m_loopPos>=1.0) {
			m_loopStartTime=time;	
		}
	}
	return changed;
}

bool Grille::triggerLoop(const LOOP_ACTION& action, const long& curTime, const LOOP_TRIGGER_MODE& mode)
{
	bool recorded=false;
	switch(m_loopState) {
		case LOOP_EMPTY : { //starts recording
			if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_PLAY) {
				for(unsigned int i=0;i<PARAMS_NB;++i) {
					m_params[VJ_PARAM(i)].startLoopRecording(curTime);
				}
				m_loopState=LOOP_RECORDING;
				m_loopStartTime=curTime;
			}
		}break;
		case LOOP_RECORDING : { //stops recording, starts playing
			if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_STOP) {
				for(unsigned int i=0;i<PARAMS_NB;++i) {
					m_params[VJ_PARAM(i)].stopLoopRecording(curTime);
				}
				m_loopState=LOOP_PLAYING;
				m_loopDuration=curTime-m_loopStartTime;
				recorded=true;
			}
		}break;
		case LOOP_OVERDUBBING : { //stop overdubbing
			if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_STOP) {
				for(unsigned int i=0;i<PARAMS_NB;++i) {
					m_params[VJ_PARAM(i)].stopLoopRecording(curTime);
				}
				m_loopState=LOOP_PLAYING;
				recorded=true;
			}
		}break;
		case LOOP_PLAYING : { //stop playing
			if(mode==LOOP_TRIGGER_OVERDUB) { //OVERDUB
				if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_PLAY) {
					for(unsigned int i=0;i<PARAMS_NB;++i) {
						m_params[VJ_PARAM(i)].startLoopRecording(curTime);
					}
					m_loopState=LOOP_OVERDUBBING;
				}
			}
			else { //NORMAL
				if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_STOP) {
					m_loopState=LOOP_STOPPED;
					m_active=false;
					m_loopPos=0;
				}
			}
			recorded=true;
		}break;
		case LOOP_STOPPED : { //restart playing
			if(action==LOOP_ACTION_TRIGGER || action==LOOP_ACTION_PLAY) {
				m_active=true;
				for(unsigned int i=0;i<PARAMS_NB;++i) {
					m_params[VJ_PARAM(i)].startLoopPlaying(curTime);
				}
				m_loopState=LOOP_PLAYING;
				m_loopPos=0;
				recorded=true;
			}
		}break;
	}	
	return recorded;
}

void Grille::eraseLoop()
{
	for(unsigned int i=0;i<PARAMS_NB;++i) {
		m_params[VJ_PARAM(i)].eraseLoop();
	}

	m_loopState=LOOP_EMPTY;
	m_loopPos=0;
}

void Grille::setLoopSpeed(const float& speed)
{
	for(unsigned int i=0;i<PARAMS_NB;++i) {
		m_params[VJ_PARAM(i)].setSpeed(speed);
	}
	m_speed=speed;
}

const float& Grille::getLoopPos()
{
	return m_loopPos;
}

void Grille::setAudioActive(const bool& act)
{
	m_audioActive = act;
}

void Grille::reinit()
{
	m_params[PARAM_ROWS].setValue(m_params[PARAM_ROWS].getInitValue());
	m_params[PARAM_COLS].setValue(m_params[PARAM_COLS].getInitValue());
	for(unsigned int i=2;i<PARAMS_NB;++i) {
		m_params[VJ_PARAM(i)].reinitPattern();
		m_params[VJ_PARAM(i)].setValue(m_params[VJ_PARAM(i)].getInitValue());
	}
}

