/***************************************************************************
 *          Parser.cpp
 *
 *  2006 Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "Parser.hpp"
#include "Vjpirate.hpp"
#include <iostream>
#include <cstdio>
#include <cstdlib>
using namespace std;

Parser::Parser(Vjpirate *t):m_vj(t)
{
	LIBXML_TEST_VERSION
	m_doc=NULL;
	m_rootElement=NULL;
}

Parser::~Parser()
{
	xmlCleanupParser();
}


int Parser::createTree(const char* filename)
{
	m_fileName=filename;
   	m_doc= xmlReadFile(filename,NULL,0);
    	m_rootElement = xmlDocGetRootElement(m_doc);
	if (m_doc == NULL || m_rootElement==NULL ) 
	{
		std::string fn(filename);
        	m_vj->showMessage("Failed to parse "+fn);
		return -1;
    	}

	if (xmlStrcmp(m_rootElement->name, (const xmlChar *) "Vjpirate")) 
	{
		m_vj->showMessage("Document of the wrong type, root node != Vjpirate");
		xmlFreeDoc(m_doc);
		return -1;
	}

	return 0;
}


int Parser::parseTree()
{
	int res = recParseTree(m_rootElement);
    	xmlFreeDoc(m_doc);
	return res;
}

int Parser::parseTreeInputs()
{
    	xmlNodePtr currentNode = NULL;
	for (currentNode = m_rootElement->children; currentNode; currentNode = currentNode->next) {
        	if (currentNode->type == XML_ELEMENT_NODE) {
			std::string strNodeName((const char*)currentNode->name);
			if(strNodeName.compare("Inputs")==0) {
				if(parseInputs(currentNode)) {
					return -1;
				}
			}
		}
	}

    	xmlFreeDoc(m_doc);
	return 0;
}

int Parser::recParseTree(xmlNode *rootNode) {
    	xmlNodePtr currentNode = NULL;
	
	for (currentNode = rootNode->children; currentNode; currentNode = currentNode->next) {
        	if (currentNode->type == XML_ELEMENT_NODE) {
			std::string strNodeName((const char*)currentNode->name);
			if(strNodeName.compare("Inputs")==0) {
				if(parseInputs(currentNode)) {
					return -1;
				}
			}
			else if(strNodeName.compare("Grilles")==0) {
				if(parseGrilles(currentNode)) {
					return -1;
				}
			}
        	}
    	}
	return 0;
}

int Parser::parseGrilles(xmlNode* grillesNode)
{
    	xmlNodePtr grilleNode = NULL;

	for (grilleNode = grillesNode->children; grilleNode; grilleNode = grilleNode->next) {
		char* value=NULL;
		float nbCols=1;    
		float nbRows=1;    
		bool homothety=true, blend=false, audio=false;
		std::string name="grille";
		GRILLE_ORIGIN origin=ORIGIN_UP_LEFT;
        	if (grilleNode->type == XML_ELEMENT_NODE) {
			std::string strNodeName((const char*)grilleNode->name);
			
			if(strNodeName.compare("Grille")==0) {

				//add a Grille to VjPirate
				Grille &gri = m_vj->addGrille();

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"name");
				if(value) {
					name=value;
				}
				gri.setName(name);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"nbCols");
				if(value) {
					nbCols=atof(value);
				}
				gri.setParameterInitValue(PARAM_COLS,nbCols/Grille::MAX_COLS);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"nbRows");
				if(value) {
					nbRows=atof(value);
				}
				gri.setParameterInitValue(PARAM_ROWS,nbRows/Grille::MAX_ROWS);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"homothety");
				if(value) {
					if(std::string(value).compare("false")==0) {
						homothety=false;
					}
				}
				gri.setHomothety(homothety);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"blend");
				if(value) {
					if(std::string(value).compare("true")==0) {
						blend=true;
					}
				}
				gri.setBlend(blend);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"audio");
				if(value) {
					if(std::string(value).compare("true")==0) {
						audio=true;
					}
				}
				gri.setAudioActive(audio);

				value=NULL;
				value = (char*)xmlGetProp(grilleNode,(xmlChar*)"origin");
				if(value) {
					if(std::string(value).compare("up-left")==0) {
						origin=ORIGIN_UP_LEFT;
					}
					else if(std::string(value).compare("up-right")==0) {
						origin=ORIGIN_UP_RIGHT;
					}
					else if(std::string(value).compare("down-right")==0) {
						origin=ORIGIN_DOWN_RIGHT;
					}
					else if(std::string(value).compare("down-left")==0) {
						origin=ORIGIN_DOWN_LEFT;
					}
				}
				gri.setOrigin(origin);

				if(parseParameters(grilleNode, gri)) {
					return -1;
				}
			}
		}
	}
	return 0;
}

int Parser::parseParameters(xmlNode* grilleNode ,Grille &gri)
{
    	xmlNodePtr paramNode = NULL;
	
	for(paramNode=grilleNode->children; paramNode; paramNode = paramNode->next) {
        	if (paramNode->type == XML_ELEMENT_NODE) {
			std::string strNodeName((const char*)paramNode->name);

			if(strNodeName.compare("Parameter")==0) {
				char* value=NULL;
				float initVal=0;    
				VJ_PARAM param=PARAM_SIZE;    
				std::string pattern="A";

				//which parameter
				value=NULL;
				value = (char*)xmlGetProp(paramNode,(xmlChar*)"type");
				if(value) {
					if(std::string(value).compare("image")==0) {
						param=PARAM_IMAGE;
					}
					else if(std::string(value).compare("size")==0) {
						param=PARAM_SIZE;
					}
					else if(std::string(value).compare("color")==0) {
						param=PARAM_COLOR;
					}
					else if(std::string(value).compare("rotation")==0) {
						param=PARAM_ROTATION;
					}
					else if(std::string(value).compare("transparency")==0) {
						param=PARAM_TRANS;
					}
					else if(std::string(value).compare("displacementx")==0) {
						param=PARAM_DISX;
					}
					else if(std::string(value).compare("displacementy")==0) {
						param=PARAM_DISY;
					}
				}

				//init value
				value=NULL;
				value = (char*)xmlGetProp(paramNode,(xmlChar*)"value");
				if(value) {
					initVal=atof(value);
				}

				
				//pattern
				value=NULL;
				value = (char*)xmlGetProp(paramNode,(xmlChar*)"pattern");
				if(value) {
					pattern=value;
				}

				//pattern elements
				
				xmlNodePtr elementNode = NULL;
				
				int elementCount=0;
				for(elementNode=paramNode->children; elementNode; elementNode = elementNode->next) {
					if (elementNode->type == XML_ELEMENT_NODE) {
						std::string strElName((const char*)elementNode->name);
						if(strElName.compare("Element")==0) {
							switch(param) {
								case PARAM_IMAGE: {
									std::vector<std::string> anims;
									float min=0,max=1;
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"anim");
									if(value) {
										anims.push_back(std::string(value));
									}
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"anim_rel");
									if(value) {
										std::string relDir = value;
										size_t pos = m_fileName.find_last_of('/');
										std::string fileDir="";
										if(pos!=string::npos) {
											fileDir=m_fileName.substr(0,pos);
										}
										relDir=fileDir+"/"+relDir;
										anims.push_back(relDir);
									}
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"min");
									if(value) {
										min=atof(value);
									}
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"max");
									if(value) {
										max=atof(value);
									}
									
									if((int)gri.getParameter(param).getNbElements()==elementCount) {
										gri.getParameter(param).addElement();
									}
									gri.setElementAnim(elementCount,m_vj->openAnim(anims));
									gri.getParameter(param).setElementRange(elementCount,min,max);
									
								}break;
								case PARAM_COLOR: {
									std::string minColor="0;0;0",maxColor="255;255;255";
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"minColor");
									if(value) {
										minColor=value;
									}
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"maxColor");
									if(value) {
										maxColor=value;
									}
									if(gri.getParameter(param).getNbElements()==elementCount) {
										gri.getParameter(param).addElement();
									}
									gri.getParameter(param).setElementColorRange(elementCount,minColor,maxColor);
								}break;
								default: {
									float min=0,max=1;
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"min");
									if(value) {
										min=atof(value);
									}
									value=NULL;
									value = (char*)xmlGetProp(elementNode,(xmlChar*)"max");
									if(value) {
										max=atof(value);
									}
									
									if(gri.getParameter(param).getNbElements()==elementCount) {
										gri.getParameter(param).addElement();
									}
									gri.getParameter(param).setElementRange(elementCount,min,max);
								}break;

							}
							
							elementCount++;

						}
					}
				}

				//set the pattern and value after having defined the elements	
				gri.getParameter(param).setInitPattern(pattern);
				gri.getParameter(param).reinitPattern();
				gri.setParameterInitValue(param,initVal);
			}

		}
	}


	return 0;
}

int Parser::parseInputs(xmlNode* inputsNode)
{
	char* value=NULL;

	int port=7000;
	//get the osc input port
	value = (char*)xmlGetProp(inputsNode,(xmlChar*)"osc_input_port");
	if(value) {
		port=atoi(value);
	}
	//TODO m_vj->setOscInputPort(port);

    	xmlNodePtr inputNode = NULL;
	//get the grilles inputs and loops inputs
	for (inputNode = inputsNode->children; inputNode; inputNode = inputNode->next) {
        	if (inputNode->type == XML_ELEMENT_NODE) {
			std::string strNodeName((const char*)inputNode->name);
			if(strNodeName.compare("Input")==0) {

				Input newInput;
				newInput.group = GRILLE_INPUT;
				newInput.type = INPUT_MIDI_CC;
				newInput.grille=0;
				newInput.parameter=0;
				newInput.midiChannel=0;
				newInput.midiCC=0;
				newInput.midiNote=0;
				newInput.midiNoteMax=0;
				newInput.audioFeature=ANALYSIS_RMS;
				newInput.oscMessage="message";
				newInput.mode=MODE_NORMAL;
				newInput.step=0.1;


				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"group");
				if(value) {
					if(std::string(value).compare("loop")==0) {
						newInput.group=LOOP_INPUT;
					}
					else if(std::string(value).compare("application")==0) {
						newInput.group=APPLICATION_INPUT;
					}
				}
				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"grille_id");
				if(value) {
					newInput.grille=atoi(value);
				}
				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"param");
				if(value) {
					std::string paramStr(value);
					Parameter p;
					p.getIdFromName(paramStr, newInput.parameter);
				}
				else { //old version
					value=NULL;					
					value = (char*)xmlGetProp(inputNode,(xmlChar*)"param_id");
					if(value) {
						newInput.parameter=atoi(value);
						if(newInput.parameter>11) {
							newInput.parameter+=4;
						}
						else if(newInput.parameter>6) {
							newInput.parameter+=2;
						}
					}
				}
				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"type");
				if(value) {
					std::string typeStr = value;
					if(typeStr.compare("midi_cc")==0) {
						newInput.type=INPUT_MIDI_CC;
					}
					else if (typeStr.compare("midi_note")==0) {
						newInput.type=INPUT_MIDI_NOTE;
					}
					else if (typeStr.compare("midi_note_range")==0) {
						newInput.type=INPUT_MIDI_NOTE_RANGE;
					}
					else if (typeStr.compare("osc")==0) {
						newInput.type=INPUT_OSC;
					}
					else if (typeStr.compare("audio")==0) {
						newInput.type=INPUT_AUDIO;
					}
					else if (typeStr.compare("keyboard")==0) {
						newInput.type=INPUT_KEYBOARD;
					}
				}

				switch(newInput.type) {
					case INPUT_MIDI_CC :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"cc");
						if(value) {
							newInput.midiCC=atoi(value);
						}
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"cc_chan");
						if(value) {
							newInput.midiChannel=atoi(value);
						}
					}break;
					case INPUT_MIDI_NOTE_RANGE :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"note_range_chan");
						if(value) {
							newInput.midiChannel=atoi(value);
						}
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"note_range_min");
						if(value) {
							newInput.midiNote=atoi(value);
						}
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"note_range_max");
						if(value) {
							newInput.midiNoteMax=atoi(value);
						}
					}break;
					case INPUT_MIDI_NOTE :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"note");
						if(value) {
							newInput.midiNote=atoi(value);
						}
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"note_chan");
						if(value) {
							newInput.midiChannel=atoi(value);
						}
					}break;
					case INPUT_OSC :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"osc_string");
						if(value) {
							newInput.oscMessage=value;
						}
					}break;
					case INPUT_KEYBOARD :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"key");
						if(value) {
							std::string keyStr(value);
							newInput.key=keyStr[0];
						}
					}break;
					case INPUT_AUDIO :{
						value=NULL;					
						value = (char*)xmlGetProp(inputNode,(xmlChar*)"audio_feature");
						if(value) {
							newInput.audioFeature=ANALYSIS_FEATURE(atoi(value)%FEATURES_NB);
						}
					}break;
				}

				//get the mode
				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"mode");
				if(value) {
					std::string modeStr(value);
					if(modeStr.compare("normal")==0) {
						newInput.mode=MODE_NORMAL;
					}
					else if(modeStr.compare("toggle")==0) {
						newInput.mode=MODE_TOGGLE;
					}
					else if(modeStr.compare("step_normal")==0) {
						newInput.mode=MODE_STEP_NORMAL;
					}
					else if(modeStr.compare("step_circular")==0) {
						newInput.mode=MODE_STEP_CIRCULAR;
					}
					else if(modeStr.compare("step_ping")==0) {
						newInput.mode=MODE_STEP_PING;
					}
					else if(modeStr.compare("step_random")==0) {
						newInput.mode=MODE_STEP_RANDOM;
					}
				}

				//get the step
				value=NULL;					
				value = (char*)xmlGetProp(inputNode,(xmlChar*)"step");
				if(value) {
					newInput.step=atof(value);
				}
				
				//add the input
				m_vj->addInput(newInput);
			}
		}
	}

	return 0;
}

int Parser::saveToFile(const char* filename)
{

	char buf[32];
	
	xmlDocPtr doc = NULL;  
    	xmlNodePtr rootNode = NULL, node1 = NULL;

	doc = xmlNewDoc(BAD_CAST "1.0");
    	rootNode = xmlNewNode(NULL, BAD_CAST "Vjpirate");
    	xmlDocSetRootElement(doc, rootNode);
	
	xmlNodePtr grillesNode = xmlNewChild(rootNode, NULL, BAD_CAST "Grilles",NULL);

	//grilles nodes	
	for(int i=0;i<m_vj->getNbGrilles();++i)
	{
		Grille& gri = m_vj->getGrille(i);
		xmlNodePtr grilleNode = xmlNewChild(grillesNode, NULL, BAD_CAST "Grille",NULL);
		xmlNewProp(grilleNode, BAD_CAST "name", BAD_CAST gri.getName().c_str());
		ostringstream oss;
		oss<<int(gri.getParameter(PARAM_COLS).getInitValue()*Grille::MAX_COLS);
		xmlNewProp(grilleNode, BAD_CAST "nbCols", BAD_CAST oss.str().c_str());
		oss.str("");
		oss<<int(gri.getParameter(PARAM_ROWS).getInitValue()*Grille::MAX_ROWS);
		xmlNewProp(grilleNode, BAD_CAST "nbRows", BAD_CAST oss.str().c_str());
		if(gri.getHomothety()) {
			xmlNewProp(grilleNode, BAD_CAST "homothety", BAD_CAST "true");
		}
		else {
			xmlNewProp(grilleNode, BAD_CAST "homothety", BAD_CAST "false");
		}
		if(gri.getBlend()) {
			xmlNewProp(grilleNode, BAD_CAST "blend", BAD_CAST "true");
		}
		else {
			xmlNewProp(grilleNode, BAD_CAST "blend", BAD_CAST "false");
		}
		if(gri.getAudioActive()) {
			xmlNewProp(grilleNode, BAD_CAST "audio", BAD_CAST "true");
		}
		else {
			xmlNewProp(grilleNode, BAD_CAST "audio", BAD_CAST "false");
		}
		switch(gri.getOrigin()) {
			case ORIGIN_UP_LEFT: xmlNewProp(grilleNode, BAD_CAST "origin", BAD_CAST "up-left"); break;
			case ORIGIN_UP_RIGHT: xmlNewProp(grilleNode, BAD_CAST "origin", BAD_CAST "up-right"); break;
			case ORIGIN_DOWN_RIGHT: xmlNewProp(grilleNode, BAD_CAST "origin", BAD_CAST "down-right"); break;
			case ORIGIN_DOWN_LEFT: xmlNewProp(grilleNode, BAD_CAST "origin", BAD_CAST "down-left"); break;
            default:break;
		}

		for(int j=2;j<PARAMS_NB;++j) {
			xmlNodePtr paramNode = xmlNewChild(grilleNode, NULL, BAD_CAST "Parameter",NULL);
			Parameter& param = gri.getParameter(VJ_PARAM(j));
			switch(VJ_PARAM(j)) {
				case PARAM_SIZE:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "size"); break;
				case PARAM_COLOR:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "color"); break;
				case PARAM_ROTATION:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "rotation"); break;
				case PARAM_TRANS:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "transparency"); break;
				case PARAM_IMAGE:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "image"); break;
				case PARAM_DISX:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "displacementx"); break;
				case PARAM_DISY:xmlNewProp(paramNode, BAD_CAST "type", BAD_CAST "displacementy"); break;
                default:break;
			}
			xmlNewProp(paramNode, BAD_CAST "pattern", BAD_CAST param.getInitPatternStr().c_str());

			oss.str("");
			oss<<param.getInitValue();
			xmlNewProp(paramNode, BAD_CAST "value", BAD_CAST oss.str().c_str());

			for(int k=0;k<param.getNbElements();++k) {
				xmlNodePtr elNode = xmlNewChild(paramNode, NULL, BAD_CAST "Element",NULL);
				Element& el = param.getElement(k);
				switch(VJ_PARAM(j)) {
					case PARAM_IMAGE: {
						if(el.getAnim()) {
							//get relative path from vjp file
							std::string animDir = el.getAnim()->getDir();
							std::string absDir, relDir;
							absDir=animDir;
							std::string fileStr = std::string (filename);
							bool equal=true;
							while(equal) {
								size_t pos1,pos2;
								std::string subFile, subAnim;
								pos1 = fileStr.find_first_of('/');
								subFile = fileStr.substr(0,pos1);
								pos2 = animDir.find_first_of('/');
								subAnim = animDir.substr(0,pos2);
								if(subFile.compare(subAnim)!=0) {	
									equal=false;
								}
								else {
									fileStr = fileStr.substr(pos1+1);
									animDir = animDir.substr(pos2+1);
								}
							}


							//test if we need to go up (../)
							size_t pos3=fileStr.find_first_of('/');
							while(pos3!=string::npos) {
								pos3 = fileStr.find_first_of('/');
								if(pos3!=string::npos) {
									fileStr=fileStr.substr(pos3+1);
									animDir="../"+animDir;
								}
							}
							
							relDir=animDir;
							//write both the abs and relative dirs
							xmlNewProp(elNode, BAD_CAST "anim", BAD_CAST absDir.c_str());
							xmlNewProp(elNode, BAD_CAST "anim_rel", BAD_CAST relDir.c_str());
						}
						else { 
							xmlNewProp(elNode, BAD_CAST "anim", BAD_CAST "");
						}
						oss.str("");
						oss<<el.getMinValue();
						xmlNewProp(elNode, BAD_CAST "min", BAD_CAST oss.str().c_str());
						oss.str("");
						oss<<el.getMaxValue();
						xmlNewProp(elNode, BAD_CAST "max", BAD_CAST oss.str().c_str());
						
					}break;
					case PARAM_COLOR: {
						oss.str("");
						oss<<el.getColorMinValue(0)<<";"<<el.getColorMinValue(1)<<";"<<el.getColorMinValue(2);
						xmlNewProp(elNode, BAD_CAST "minColor", BAD_CAST oss.str().c_str());
						oss.str("");
						oss<<el.getColorMaxValue(0)<<";"<<el.getColorMaxValue(1)<<";"<<el.getColorMaxValue(2);
						xmlNewProp(elNode, BAD_CAST "maxColor", BAD_CAST oss.str().c_str());
					}break;
					default: {
						oss.str("");
						oss<<el.getMinValue();
						xmlNewProp(elNode, BAD_CAST "min", BAD_CAST oss.str().c_str());
						oss.str("");
						oss<<el.getMaxValue();
						xmlNewProp(elNode, BAD_CAST "max", BAD_CAST oss.str().c_str());
					}break;
				}
			}

			
		}
		
	}
	

	xmlNodePtr inputsNode = xmlNewChild(rootNode, NULL, BAD_CAST "Inputs",NULL);

	//get all the grille inputs
	for(unsigned int i=0;i<m_vj->getNbInputs();++i) {

		Input& inp = m_vj->getInput(i);

		ostringstream oss;
		xmlNodePtr inpNode = xmlNewChild(inputsNode, NULL, BAD_CAST "Input",NULL);

		switch(inp.group) {
			case LOOP_INPUT: {
				oss.str("loop");
			}break;
			case GRILLE_INPUT: {
				oss.str("grille");
			}break;
			case APPLICATION_INPUT: {
				oss.str("application");
			}break;
		}

		xmlNewProp(inpNode, BAD_CAST "group", BAD_CAST oss.str().c_str());

		oss.str("");
		oss<<inp.grille;
		xmlNewProp(inpNode, BAD_CAST "grille_id", BAD_CAST oss.str().c_str());

		std::string paramStr;
		Parameter p;
		p.getNameFromId(inp.parameter, paramStr);
		xmlNewProp(inpNode, BAD_CAST "param", BAD_CAST paramStr.c_str());

		switch(inp.type) {
			case INPUT_MIDI_NOTE: {
				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "midi_note");

				oss.str("");
				oss<<inp.midiChannel;
				xmlNewProp(inpNode, BAD_CAST "note_chan", BAD_CAST oss.str().c_str());

				oss.str("");
				oss<<inp.midiNote;
				xmlNewProp(inpNode, BAD_CAST "note", BAD_CAST oss.str().c_str());
			}break;
			case INPUT_MIDI_NOTE_RANGE: {
				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "midi_note_range");

				oss.str("");
				oss<<inp.midiChannel;
				xmlNewProp(inpNode, BAD_CAST "note_range_chan", BAD_CAST oss.str().c_str());

				oss.str("");
				oss<<inp.midiNote;
				xmlNewProp(inpNode, BAD_CAST "note_range_min", BAD_CAST oss.str().c_str());

				oss.str("");
				oss<<inp.midiNoteMax;
				xmlNewProp(inpNode, BAD_CAST "note_range_max", BAD_CAST oss.str().c_str());
			}break;
			case INPUT_MIDI_CC: {

				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "midi_cc");

				oss.str("");
				oss<<inp.midiChannel;
				xmlNewProp(inpNode, BAD_CAST "cc_chan", BAD_CAST oss.str().c_str());

				oss.str("");
				oss<<inp.midiCC;
				xmlNewProp(inpNode, BAD_CAST "cc", BAD_CAST oss.str().c_str());
			}break;
			case INPUT_OSC: {

				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "osc");

				oss.str("");
				oss<<inp.oscMessage;
				xmlNewProp(inpNode, BAD_CAST "osc_string", BAD_CAST oss.str().c_str());
			}break;
			case INPUT_AUDIO: {

				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "audio");

				oss.str("");
				oss<<inp.audioFeature;
				xmlNewProp(inpNode, BAD_CAST "audio_feature", BAD_CAST oss.str().c_str());
			}break;
			case INPUT_KEYBOARD: {
				xmlNewProp(inpNode, BAD_CAST "type", BAD_CAST "keyboard");

				oss.str("");
				oss<<inp.key;
				xmlNewProp(inpNode, BAD_CAST "key", BAD_CAST oss.str().c_str());
			}break;
		}

		//trigger mode
		switch(inp.mode) {
			case MODE_NORMAL: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "normal");
			}break;
			case MODE_TOGGLE: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "toggle");
			}break;
			case MODE_STEP_NORMAL: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "step_normal");
			}break;
			case MODE_STEP_CIRCULAR: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "step_circular");
			}break;
			case MODE_STEP_PING: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "step_ping");
			}break;
			case MODE_STEP_RANDOM: {
				xmlNewProp(inpNode, BAD_CAST "mode", BAD_CAST "step_random");
			}break;
		}

		//trigger step
		oss.str("");
		oss<<inp.step;
		xmlNewProp(inpNode, BAD_CAST "step", BAD_CAST oss.str().c_str());
		
	}


	xmlSaveFormatFileEnc(filename, doc, "UTF-8", 1);

	xmlFreeDoc(doc);

	xmlCleanupParser();
	return 0;
}

