/***************************************************************************
 *            Vjpirate.cpp
 *
 *  2005  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "Vjpirate.hpp"
#include <math.h>
#include "Parser.hpp"
#include <sstream>
using namespace std;

Vjpirate::Vjpirate(): m_selectActive(true), m_loopEraseToggle(false)  {}

Vjpirate::~Vjpirate(){}

void Vjpirate::start() {
	m_fileName="";

	clearInputs();

	//set the loop number to 8 //TODO dynamic number of loops
	setNbLoops(NB_LOOPS);

	//start the ui
	m_execWin = new ExecWindow(900,700,"VjPirate",this);	
	m_execWin->show();

	//create a grille
	addGrille();
	m_currentGrille=0;

	m_execWin->reset();

	//input manager
	m_inputManager = InputManager::getInstance();	
	m_inputManager->start();
    cout<<"VjPirate started"<<endl;
}

void Vjpirate::stop() {
	showMessage("Stopping input");
	if(m_inputManager) {
		m_inputManager->stop();
	}
	//remove anims
	clearAllAnims();
}


Vjpirate* Vjpirate::getInstance()
{
    static Vjpirate instance;
    return &instance;
}


void Vjpirate::update() {

    m_inputManager->process();

	int selLoop=Vjpirate::SELECTED_LOOP;
	timeval curTime;
    gettimeofday(&curTime,NULL);
    long longCurTime = (long)(curTime.tv_sec-((int)((float)curTime.tv_sec/1000000.0)*1000000))*1000 + (long)((float)curTime.tv_usec/1000.0);
    float loopPos;
    getCurrentGrille().playLoop(longCurTime);
    if((loopPos=getCurrentGrille().getLoopPos())>=0) {
        m_execWin->setLoopPos(selLoop, loopPos);
    }
    for(unsigned int loop=0; loop<getNbLoops(); ++loop) {
        getLoop(loop).playLoop(longCurTime);
        if((loopPos=getLoop(loop).getLoopPos())>=0) {
            m_execWin->setLoopPos(loop, loopPos);
        }
    }

}

void Vjpirate::showMessage(std::string mess)
{
	std::string copiedMess=mess;
	cout<<"<VjPirate> "<<copiedMess<<endl;
}

void Vjpirate::openFile(const char* fn, bool wait)
{
	if(wait) {
		std::string fileName(fn);
		m_execWin->setOpenfile(fileName);
	}
	else {
		Parser pars(this);
		if(!pars.createTree(fn)) {
			//clear the existing grilles
			clearGrilles();
			//clear loops
			clearLoops();
			//clear the existing inputs
			clearInputs();
			//clear anims
			clearAllAnims();
			//parse the file
			if(pars.parseTree()) {
				showMessage("Error Parsing "+std::string(fn));
			}
			else {
				m_fileName=fn;
				m_execWin->reset();
				std::string title = "VjPirate: ";
				title+=getShortFileName();
				m_execWin->setTitle(title.c_str());
				setCurrentGrille(0);
				resetAllInputs();
			}
		}
	}
}

void Vjpirate::saveToFile(const char* filename)
{
	Parser pars(this);
	pars.saveToFile(filename);
	m_fileName=filename;
}

void Vjpirate::importInputs(const char* fn)
{
	Parser pars(this);
	if(!pars.createTree(fn)) {
		//clear the existing inputs
		clearInputs();
		//parse the file
		if(pars.parseTreeInputs()) {
			showMessage("Error Parsing "+std::string(fn));
		}
		//reset the inputs
		resetAllInputs();
	}
}

void Vjpirate::clearGrilles()
{
	m_grilles.clear();
}

void Vjpirate::clearLoops()
{
	for(unsigned int i=0;i<m_loops.size();++i) {
		m_loops[i] = Grille();	
	}	
}

void Vjpirate::swapGrilles(int gri1, int gri2) {
	if(gri1<(int)m_grilles.size() 
            && gri2<(int)m_grilles.size() 
            && gri1>=0 && gri2>=0) {
		Grille tempGrille;
		tempGrille=m_grilles[gri1];
		m_grilles[gri1]=m_grilles[gri2];
		m_grilles[gri2]=tempGrille;
		
		
		//swap inputs
		for(unsigned int i=0;i<m_inputs.size();++i) {
			if(m_inputs[i].group==GRILLE_INPUT) {
				if(m_inputs[i].grille==gri1+1) {
					m_inputs[i].grille=gri2+1;
				}
				else if(m_inputs[i].grille==gri2+1) {
					m_inputs[i].grille=gri1+1;
				}
			}
		}

		resetAllInputs();

		//redraw
		m_execWin->resetExecGrilles();
	}
}

Grille& Vjpirate::addGrille()
{
	m_grilles.push_back(Grille());

	m_execWin->resetExecGrilles();
	m_execWin->setSelectedExecGrille(m_currentGrille);

	showMessage("Added a grille");

	return m_grilles[m_grilles.size()-1];
}

void Vjpirate::removeGrille(int gri) {
	if(gri<(int)m_grilles.size() && m_grilles.size()>1) {
		//m_grilles[gri].removeAnims();
		m_execWin->reinitGl();
		m_grilles.erase(m_grilles.begin()+gri);

		for(unsigned int i=0;i<m_inputs.size();++i) {
			if(m_inputs[i].group==GRILLE_INPUT) {
				if(m_inputs[i].grille==gri+1) {
					m_inputs.erase(m_inputs.begin()+i);
					i--;
				}
				else {
					if(m_inputs[i].grille>gri+1) {
						m_inputs[i].grille = m_inputs[i].grille-1;
					}
				}
			}
		}


		if(m_currentGrille>=(int)m_grilles.size()) {
			setCurrentGrille(m_grilles.size()-1);
		}
		
		resetAllInputs();
		m_execWin->resetExecGrilles();
		m_execWin->setSelectedExecGrille(m_currentGrille);

		showMessage("Removed a grille");
	}
}

void Vjpirate::renameGrille(int gri, std::string name)
{
	m_grilles[gri].setName(name);	
	m_execWin->resetInputsTab();
}

Grille& Vjpirate::duplicateGrille(int gri)
{
	if(gri<(int)m_grilles.size()) {
		m_grilles.push_back(Grille());
		m_grilles[m_grilles.size()-1].copy(m_grilles[gri]);
		m_execWin->resetExecGrilles();
		m_execWin->setSelectedExecGrille(m_currentGrille);
		showMessage("Duplicated a grille");
	}
	return m_grilles[m_grilles.size()-1];
}

void Vjpirate::resetGrillesGlCoords()
{
	for(unsigned int i=0;i<m_grilles.size();++i) {
		m_grilles[i].hasGlCoords(false);
	}	
	for(unsigned int i=0;i<m_loops.size();++i) {
		m_loops[i].hasGlCoords(false);
	}	

}

Anim* Vjpirate::openAnim(std::vector<std::string> dirs)
{
	Anim* resAnim=NULL;
	bool opened=false;
	for(unsigned int i=0;i<dirs.size() && opened==false;++i) {
		if((resAnim=openAnim(dirs[i]))!=NULL) {
			opened=true;
		}
	}

	return resAnim;
}

Anim* Vjpirate::openAnim(std::string dir)
{
	Anim* resAnim=NULL;

	if(dir.compare("")!=0) {
		//test if already opened	
		std::vector<Anim*>::iterator it;
		for(it=m_anims.begin();it!=m_anims.end();++it) {
			if((*it)->getDir().compare(dir)==0) {
				(*it)->addUser();
				resAnim=(*it);
				resAnim->setAction(ANIM_ACTION_LOAD);
				m_execWin->reinitGl();
			}
		}
		if(!resAnim) {	
			showMessage("Opening Anim: "+dir);
			Anim* newAnim = new Anim(dir);
			if(newAnim->testDir()) {
				m_anims.push_back(newAnim);	
				resAnim=newAnim;
				resAnim->parseImages();
				resAnim->setAction(ANIM_ACTION_LOAD);
				m_execWin->reinitGl();
			}
			else {
				showMessage("Error loading anim "+dir);
				delete newAnim;	
			}
		}
	}
	return resAnim;
}


void Vjpirate::closeAnim(std::string dir)
{
	std::vector<Anim*>::iterator it;
	for(it=m_anims.begin();it!=m_anims.end();++it) {
		if((*it)->getDir().compare(dir)==0) {
			if((*it)->removeUser()<=0) {
				Anim* tmpAnim = (*it);
				tmpAnim->setAction(ANIM_ACTION_UNLOAD);
			}
			break;
		}
	}
	m_execWin->reinitGl();
}

void Vjpirate::testAllAnims()
{
	//showMessage("Loading all anims");
	std::vector<Anim*>::iterator it;
	for(it=m_anims.begin();it!=m_anims.end();it++) {
		Anim* an = *it;
		an->testAction();
		if(!an->getLoaded()) {
			m_anims.erase(it);
			delete an;
			it--;
		}
	}
}

void Vjpirate::clearAllAnims()
{
	std::vector<Anim*>::iterator it;
	for(it=m_anims.begin();it!=m_anims.end();it++) {
		Anim* tmpAnim = (*it);
		tmpAnim->setAction(ANIM_ACTION_UNLOAD);
	}
	m_execWin->reinitGl();
}

GLuint Vjpirate::getNewTextureID()
{
	GLuint res=0;
	int pos=-1;
	for(unsigned int i=0;i<m_textureIDs.size() && pos<0;++i) {
		if(!m_textureIDUsed[i]) {
			pos=i;
		}
	}

	if(pos>=0) {
		res=m_textureIDs[pos];
		m_textureIDUsed[pos]=true;
	}
	else {
		GLuint newID;
		glGenTextures( 1, &newID);
		m_textureIDs.push_back(newID);
		m_textureIDUsed.push_back(true);
		res=newID;
	}

	return res;
}

void Vjpirate::freeTextureID(GLuint id)
{
	bool found=false;
	for(unsigned int i=0;i<m_textureIDs.size() && !found;++i) {
		if(m_textureIDs[i]==id) {
			found=true;
			glDeleteTextures( 1, &id);
			m_textureIDUsed[i]=false;
		}
	}
}

bool Vjpirate::isParameterofCurrentGrilleControlledByAudio(int param)
{
	bool res=false;
	for(unsigned int i=0;i<m_inputs.size();++i) {
		if(m_inputs[i].group==GRILLE_INPUT && (m_inputs[i].grille==m_currentGrille+1 || m_inputs[i].grille==SELECTED_GRILLE)) {
			if(m_inputs[i].parameter==param && m_inputs[i].type==INPUT_AUDIO) {
				res=true;
			}
		}
	}

	return res;
}

bool Vjpirate::isParameterPermutofCurrentGrilleControlledByAudio(int param)
{
	bool res=false;
	for(unsigned int i=0;i<m_inputs.size();++i) {
		if(m_inputs[i].group==GRILLE_INPUT && (m_inputs[i].grille==m_currentGrille+1 || m_inputs[i].grille==SELECTED_GRILLE)) {
			if(m_inputs[i].parameter==param-2+PARAMS_NB && m_inputs[i].type==INPUT_AUDIO) {
				res=true;
			}
		}
	}

	return res;
}

void Vjpirate::processAnalysis(double* features) {

	//test if need to reset the inputs
	if(m_resetInputsAsked[INPUT_AUDIO]) {
		resetInputs(INPUT_AUDIO, false);
		m_resetInputsAsked[INPUT_AUDIO]=false;
	}


	for(unsigned int i=0;i<FEATURES_NB;++i) {
		//test grilles audio inputs
		for(unsigned int j=0;j<m_audioInputs[i].size();++j) {
			Input& inp = m_inputs[m_audioInputs[i][j]];

			float newVal=0;
			inp.processValue(features[i], newVal);

			if(inp.group==GRILLE_INPUT) {
				int grilleID = inp.grille;
				int paramID = inp.parameter;
				if(grilleID==SELECTED_GRILLE) { //selected grille
					if(paramID<PARAMS_NB) { //parameters
						if(m_grilles[m_currentGrille].getAudioActive()) {
							m_grilles[m_currentGrille].getParameter(VJ_PARAM(paramID)).setValue(newVal, AUDIO_VALUE);
							m_execWin->setCurrentGrilleParameterValue(paramID,newVal);
						}
					}
					else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB) { //permutations
						if(m_grilles[m_currentGrille].getAudioActive()) {
							m_grilles[m_currentGrille].getParameter(VJ_PARAM(paramID-PARAMS_NB+2)).setPatternPermutation(newVal, AUDIO_VALUE);
							m_execWin->setCurrentGrilleParameterPatternPermutation(paramID-PARAMS_NB+2,newVal);
						}
					}
				}
				else {
					if(grilleID-1<m_grilles.size()) {
						if(paramID<PARAMS_NB) { //parameters
							if(m_grilles[grilleID-1].getAudioActive()) {
								m_grilles[grilleID-1].getParameter(VJ_PARAM(paramID)).setValue(newVal, AUDIO_VALUE);
								if(grilleID-1==m_currentGrille) {
									m_execWin->setCurrentGrilleParameterValue(paramID,newVal);
								}
							}
						}
						else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB) { //permutations
							if(m_grilles[grilleID-1].getAudioActive()) {
								m_grilles[grilleID-1].getParameter(VJ_PARAM(paramID-PARAMS_NB+2)).setPatternPermutation(newVal, AUDIO_VALUE);
								if(grilleID-1==m_currentGrille) {
									m_execWin->setCurrentGrilleParameterPatternPermutation(paramID-PARAMS_NB+2,newVal);
								}
							}
						}
					}
				}
			}
		}

		//test loops audio inputs
		for(unsigned int j=0;j<m_loopAudioInputs[i].size();++j) {
			Input& inp = m_loopAudioInputs[i][j];
			float newVal=0;
			inp.processValue(features[i], newVal);
			int loopID = inp.grille;
			int paramID = inp.parameter;
			if(paramID<PARAMS_NB) { 
				m_loops[loopID].getParameter(VJ_PARAM(paramID)).setValue(newVal, AUDIO_VALUE);
			}
			else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB) { //permutations
				m_loops[loopID].getParameter(VJ_PARAM(paramID-PARAMS_NB+2)).setPatternPermutation(newVal, AUDIO_VALUE);
			}
		}
	}

}
	
void Vjpirate::processInputEvent(const INPUT_TYPE& inputType)
{
	//test if need to reset the inputs
	if(m_resetInputsAsked[inputType]) {
		resetInputs(inputType, false);
		m_resetInputsAsked[inputType]=false;
	}

	std::vector<int> resInputs;
	bool found=false;
	bool extFound=false;

	InputEvent newEvent=m_inputEvents[inputType];

	if(m_learnInput) {

		Input newInput;
		newInput.group=m_learnInputArgs[1];
		newInput.grille=m_learnInputArgs[2];
		newInput.parameter=m_learnInputArgs[3];
		newInput.type=inputType;
		newInput.midiChannel=newEvent.midiChannel;
		newInput.oscMessage=newEvent.oscMessage;
		newInput.midiCC=newEvent.midiParam;
		newInput.midiNote=newEvent.midiParam;
		newInput.midiNoteMax=newEvent.midiParam;
		newInput.key=newEvent.key;
		newInput.mode=MODE_NORMAL;

		//if creating a new input
		if(m_learnInputArgs[0]==LEARN_INPUT_CREATE) {
			//look for an existing input with the same group/grille/param/type
			int inp=0;
			bool found=false;
			for(inp=0;inp<m_inputs.size() && !found;) {
				if(m_inputs[inp].group==m_learnInputArgs[1] && m_inputs[inp].grille==m_learnInputArgs[2] && m_inputs[inp].parameter==m_learnInputArgs[3] && m_inputs[inp].type==inputType) {
					found=true;	
				}
				else {
					++inp;
				}
			}

			if(found) {
				setInput(inp, newInput, true);
			}
			else {
				addInput(newInput);
			}
		}
		else {

			//test if learning a midi note range and receiving a note
			if(m_inputs[m_learnInputArgs[0]].type==INPUT_MIDI_NOTE_RANGE && inputType==INPUT_MIDI_NOTE) {
				newInput.type=INPUT_MIDI_NOTE_RANGE;
				setInput(m_learnInputArgs[0], newInput, true);
			}
			else {
				setInput(m_learnInputArgs[0], newInput, true);
			}
		}
		m_learnInput=false;
	}
	else {
		switch(inputType) {
			case INPUT_MIDI_CC: {
				int input = newEvent.midiChannel*1000+newEvent.midiParam;
				std::map<int, std::vector<int> >::iterator it = m_midiCCInputs.find(input);
				if(it!=m_midiCCInputs.end()) {
					for(unsigned int i=0;i<it->second.size();++i) {
						resInputs.push_back(it->second[i]);
					}
					found=true;
				}
			}break;
			case INPUT_MIDI_NOTE: {
				int input = newEvent.midiChannel*1000+newEvent.midiParam;
				std::map<int, std::vector<int> >::iterator it = m_midiNoteInputs.find(input);
				if(it!=m_midiNoteInputs.end()) {
					for(unsigned int i=0;i<it->second.size();++i) {
						resInputs.push_back(it->second[i]);
					}
					found=true;
				}
			} break;
			case INPUT_MIDI_NOTE_RANGE: {
				int input = newEvent.midiChannel*1000+newEvent.midiParam;
				std::map<int, std::vector<int> >::iterator it = m_midiNoteRangeInputs.find(input);
				if(it!=m_midiNoteRangeInputs.end()) {
					for(unsigned int i=0;i<it->second.size();++i) {
						resInputs.push_back(it->second[i]);
					}
					found=true;
				}
			} break;
			case INPUT_OSC: {
				std::map<std::string, std::vector<int> >::iterator it = m_oscInputs.find(newEvent.oscMessage);
				if(it!=m_oscInputs.end()) {
					for(unsigned int i=0;i<it->second.size();++i) {
						resInputs.push_back(it->second[i]);
					}
					found=true;
				}
			}break;
			case INPUT_KEYBOARD: {
				std::map<char, std::vector<int> >::iterator it = m_keyboardInputs.find(newEvent.key);
				if(it!=m_keyboardInputs.end()) {
					for(unsigned int i=0;i<it->second.size();++i) {
						resInputs.push_back(it->second[i]);
					}
					found=true;
				}
			}break;
		}
	}

	//if found an input
	if(found) {
		//get the grille(s) and parameter
		for(unsigned int i=0;i<resInputs.size();++i) {
			float tempVal=newEvent.value;

			//get value if input is a note range
			if(inputType==INPUT_MIDI_NOTE_RANGE) {
				int min = m_inputs[resInputs[i]].midiNote;
				int max = m_inputs[resInputs[i]].midiNoteMax;
				if(max==min) {
					tempVal=1;
				}
				else {
					tempVal=float(newEvent.midiParam-min)/float(max-min);
				}
				//cout<<param<<" "<<min<<" "<<max<<" "<<value<<endl;
			}

			//process the value
			float newVal=0;
			m_inputs[resInputs[i]].processValue(tempVal,newVal);

			if(newVal>=0) {

				//check grille inputs
				if(m_inputs[resInputs[i]].group==GRILLE_INPUT) {
					int grilleID = m_inputs[resInputs[i]].grille;
					int paramID = m_inputs[resInputs[i]].parameter;


					//cout<<"grille input "<<grilleID<<" "<<paramID<<endl;

					//if input to selected grille
					if(grilleID==SELECTED_GRILLE) { //selected grille
						if(paramID<PARAMS_NB) {
							//cout<<"input to params"<<endl;
							setCurrentGrilleParameterValue(VJ_PARAM(paramID),newVal);
							m_execWin->setCurrentGrilleParameterValue(paramID,newVal);
						}
						else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB) {
						//	cout<<"input to permuts , param "<<paramID<<endl;
							setCurrentGrilleParameterPatternPermutation(VJ_PARAM(paramID-PARAMS_NB+2),newVal);
							m_execWin->setCurrentGrilleParameterPatternPermutation(paramID-PARAMS_NB+2,newVal);
						}
						else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB+ADD_PARAMS_NB) {
						//	cout<<"input to add"<<endl;
							paramID-=(PARAMS_NB+PERMUT_PARAMS_NB);
							switch(paramID) {
								case PARAM_ACTIVATION: { //activation
									m_grilles[m_currentGrille].setActive(newVal);
									m_execWin->setCurrentGrilleAct(newVal);
									resetAudioInputStoredValues(GRILLE_INPUT, 
															m_currentGrille);
								}break;
								case PARAM_REINIT: { //reinit
									m_grilles[m_currentGrille].reinit();
									m_execWin->setSelectedExecGrille(m_currentGrille);
								}break;
								case PARAM_BLEND: { //blend
									m_grilles[m_currentGrille].setBlend(newVal);
									m_execWin->setCurrentGrilleBlend(newVal);
								}break;
								case PARAM_AUDIO: { //audio
									m_grilles[m_currentGrille].setAudioActive(newVal);
									m_execWin->setCurrentGrilleAudio(newVal);
									resetAudioInputStoredValues(GRILLE_INPUT, 
															m_currentGrille);
								}break;
							}
						}
					}
					else {
						grilleID--;
						if(grilleID<m_grilles.size()) {
							if(paramID<PARAMS_NB) {
								//cout<<"curr grille param "<<value<<endl;
								m_grilles[grilleID].setParameterValue(VJ_PARAM(paramID),newVal);
								if(grilleID==m_currentGrille) {
									m_execWin->setCurrentGrilleParameterValue(paramID,newVal);
								}
							}
							else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB) {
								m_grilles[grilleID].setParameterPatternPermutation(VJ_PARAM(paramID-PARAMS_NB+2),newVal);
								if(grilleID==m_currentGrille) {
									m_execWin->setCurrentGrilleParameterPatternPermutation(paramID-PARAMS_NB+2,newVal);
								}
							}
							else if(paramID<PARAMS_NB+PERMUT_PARAMS_NB+ADD_PARAMS_NB) {
								paramID-=(PARAMS_NB+PERMUT_PARAMS_NB);
								switch(paramID) {
									case PARAM_ACTIVATION: { //activation
										m_grilles[grilleID].setActive(newVal);
										if(grilleID==m_currentGrille) {
											m_execWin->setCurrentGrilleAct(newVal);
										}
										resetAudioInputStoredValues(
															GRILLE_INPUT, 
															grilleID);
									}break;
									case PARAM_REINIT: { //reinit
										m_grilles[grilleID].reinit();
										if(grilleID==m_currentGrille) {
											m_execWin->setSelectedExecGrille(m_currentGrille);
										}
									}break;
									case PARAM_BLEND: { //blend
										m_grilles[grilleID].setBlend(newVal);
										if(grilleID==m_currentGrille) {
											m_execWin->setCurrentGrilleBlend(newVal);
										}
									}break;
									case PARAM_AUDIO: { //audio
										m_grilles[grilleID].setAudioActive(newVal);
										if(grilleID==m_currentGrille) {
											m_execWin->setCurrentGrilleAudio(newVal);
										}
										resetAudioInputStoredValues(
															GRILLE_INPUT, 
															grilleID);
									}break;
								}
							}
						}
					}

				}
				else if(m_inputs[resInputs[i]].group==LOOP_INPUT) {

					int loopID = m_inputs[resInputs[i]].grille;
					int paramID = m_inputs[resInputs[i]].parameter;

					switch(paramID) {
						case LOOP_ERASE_TOGGLE: {
							if(newVal>0) {
								m_loopEraseToggle=true;
							}
							else {
								m_loopEraseToggle=false;
							}
						}break;
						case LOOP_ERASE: {
							if(newVal>0) {
								eraseLoop(loopID);
							}
						}break;
						case LOOP_SPEED: {
							setLoopSpeed(loopID,newVal);
						}break;
						case LOOP_TRANSPARENCY: {
							setLoopTransparency(loopID,newVal);
						}break;
						case LOOP_TRIGGER: {
							if(m_loopEraseToggle) {
								eraseLoop(loopID);
							}
							else {
								if(newVal>0) {
									triggerLoop(LOOP_ACTION_PLAY, loopID);
								}
								else {
									triggerLoop(LOOP_ACTION_STOP, loopID);
								}
							}
						}break;
					}
				}
				else if(m_inputs[resInputs[i]].group==APPLICATION_INPUT) {
					int paramID = m_inputs[resInputs[i]].parameter;
					switch(paramID) {
						case PARAM_MASTER: { //master transparency
							setMasterTransparency(newVal);
						}break;
						case PARAM_SELECT: { //selection
							setCurrentGrille(newVal*float(m_grilles.size()));
						}break;
						case PARAM_ACTIVE_SELECT: { //activate when select
							setSelectActive(newVal);
						}break;
						case PARAM_LOOP_RECORD: { //loop record/overdub
							int selLoop = SELECTED_LOOP;
							if(newVal==0) {
								triggerLoop(LOOP_ACTION_STOP, selLoop);
							}
							else {
								triggerLoop(LOOP_ACTION_PLAY, selLoop);
							}
						}break;
						case PARAM_LOOP_ERASE: { //loop erase
							if(newVal>0) {
								int selLoop = SELECTED_LOOP;
								eraseLoop(selLoop);
							}
						}break;
					}
				}
			}
		}

	}
}

void Vjpirate::setCurrentGrille(unsigned int gri)
{
	if(gri>=m_grilles.size()) {
		gri=m_grilles.size()-1; 
	}

	m_grilles[m_currentGrille].setActive(false);
	m_currentGrille=gri;	
	if(m_selectActive) {
		m_grilles[m_currentGrille].setActive(true);
	}
	m_execWin->setSelectedExecGrille(m_currentGrille);
	int selGrille = SELECTED_GRILLE;
	setInputStoredValue(GRILLE_INPUT, selGrille, int(PARAMS_NB+PERMUT_PARAMS_NB+PARAM_BLEND), m_grilles[m_currentGrille].getBlend());
	setInputStoredValue(GRILLE_INPUT, selGrille, int(PARAMS_NB+PERMUT_PARAMS_NB+PARAM_ACTIVATION), m_grilles[m_currentGrille].getActive());
}

void Vjpirate::addInput(Input &inp)
{
	//go through the inputs and put this one at the right place
	int index=0;
	bool found=false;
	for(;index<m_inputs.size() && !found;) {
		if(m_inputs[index].group>inp.group || (m_inputs[index].group==inp.group && m_inputs[index].grille>inp.grille) ||  (m_inputs[index].group==inp.group && m_inputs[index].grille==inp.grille && m_inputs[index].parameter>=inp.parameter) ) {
			found=true;
		}	
		else {
			index++;
		}
	}

	m_inputs.insert(m_inputs.begin()+index,inp);

	//activate the grille if added an audio input
	/*
	if(inp.type==INPUT_AUDIO && inp.group==GRILLE_INPUT && inp.grille!=SELECTED_GRILLE) {
		m_grilles[inp.grille-1].setAudioActive(true);	
	}*/	

	resetAllInputs();
	m_execWin->resetInputsTab();
}

void Vjpirate::addDefaultInput()
{
	Input inp;
	inp.group=GRILLE_INPUT;
	inp.grille=0;
	inp.parameter=0;
	inp.type=INPUT_MIDI_CC;
	inp.mode=MODE_NORMAL;
	inp.midiChannel=0;
	inp.midiCC=0;
	addInput(inp);
	resetAllInputs();
	m_execWin->resetInputsTab();
}

void Vjpirate::setInput(const int& id, Input &inp, const bool& learning)
{
	INPUT_TYPE oldType = m_inputs[id].type;
	m_inputs[id] = inp;
	resetInputs(inp.type);
	resetInputs(oldType);
	if(learning) {
		m_execWin->resetInputsTab();
	}
}

void Vjpirate::removeInput(int id)
{
	m_inputs.erase(m_inputs.begin()+id);
	resetAllInputs();
	m_execWin->resetInputsTab();
}

void Vjpirate::learnInput(const bool& act, const std::vector<int>& args)
{
	m_learnInputArgs=args;	
	m_learnInput=act;
}

void Vjpirate::setInputStoredValue(const INPUT_GROUP& group, const int& grille, const int& parameter, const float& value)
{
	for(unsigned int i=0;i<m_inputs.size();++i) {
		if(m_inputs[i].group==group && m_inputs[i].grille==grille && m_inputs[i].parameter==parameter) {
			m_inputs[i].prevOut=value;
		}
	}
}

void Vjpirate::resetAudioInputStoredValues(const INPUT_GROUP& group, 
													const int& grille) {
	for(unsigned int i=0;i<m_inputs.size();++i) {
		if(m_inputs[i].group==group && m_inputs[i].grille-1==grille
				&& m_inputs[i].type==INPUT_AUDIO) {
			m_inputs[i].prevOut=0;
		}
	}
}

void Vjpirate::resetAllInputs() 
{
	resetInputs(INPUT_AUDIO);	
	resetInputs(INPUT_KEYBOARD);	
	resetInputs(INPUT_OSC);	
	resetInputs(INPUT_MIDI_CC);	
	resetInputs(INPUT_MIDI_NOTE);	
	resetInputs(INPUT_MIDI_NOTE_RANGE);	
}

void Vjpirate::resetInputs(const INPUT_TYPE& type, const bool& wait) 
{

	if(wait) {
		m_resetInputsAsked[type]=true;
	}
	else {
		switch(type) {
			case INPUT_AUDIO : {
				for(unsigned int i=0;i<FEATURES_NB;++i) {
					m_audioInputs[i].clear();
				}
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type==INPUT_AUDIO){
						m_audioInputs[int(inp.audioFeature)].push_back(i);
					}
				}
			}break;
			case INPUT_OSC: {
				m_oscInputs.clear();
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type == INPUT_OSC) {
						std::map<std::string, std::vector<int> >::iterator it = m_oscInputs.find(inp.oscMessage);
						if(it!=m_oscInputs.end()) {
							it->second.push_back(i);
						}
						else {
							std::vector<int> vec(1,i);
							m_oscInputs[inp.oscMessage] = vec;
						}
					}
				}
			}break;
			case INPUT_KEYBOARD: {
				m_keyboardInputs.clear();
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type==INPUT_KEYBOARD) {
						char id = inp.key;
						std::map<char, std::vector<int> >::iterator it = m_keyboardInputs.find(id);
						if(it!=m_keyboardInputs.end()) {
							it->second.push_back(i);
						}
						else {
							std::vector<int> vec(1,i);
							m_keyboardInputs[id] = vec;
						}
					}
				}
			}break;
			case INPUT_MIDI_CC: {
				m_midiCCInputs.clear();
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type == INPUT_MIDI_CC) {
						int intID = inp.midiChannel*1000+inp.midiCC;
						std::map<int, std::vector<int> >::iterator it = m_midiCCInputs.find(intID);
						if(it!=m_midiCCInputs.end()) {
							it->second.push_back(i);
						}
						else {
							std::vector<int> vec(1,i);
							m_midiCCInputs[intID] = vec;
						}
					}
				}
			}break;
			case INPUT_MIDI_NOTE: {
				m_midiNoteInputs.clear();
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type == INPUT_MIDI_NOTE) {
						int intID = inp.midiChannel*1000+inp.midiNote;
						std::map<int, std::vector<int> >::iterator it = m_midiNoteInputs.find(intID);
						if(it!=m_midiNoteInputs.end()) {
							it->second.push_back(i);
						}
						else {
							std::vector<int> vec(1,i);
							m_midiNoteInputs[intID] = vec;
						}
					}
				}
			}break;
			case INPUT_MIDI_NOTE_RANGE: {
				m_midiNoteRangeInputs.clear();
				for(unsigned int i=0;i<m_inputs.size();++i) {
					Input &inp = m_inputs[i];
					if(inp.type==INPUT_MIDI_NOTE_RANGE) {
						for(unsigned int r=inp.midiNote; r<=inp.midiNoteMax; ++r) {
							int intID = inp.midiChannel*1000+r;
							std::map<int, std::vector<int> >::iterator it = m_midiNoteRangeInputs.find(intID);
							if(it!=m_midiNoteRangeInputs.end()) {
								it->second.push_back(i);
							}
							else {
								std::vector<int> vec(1,i);
								m_midiNoteRangeInputs[intID] = vec;
							}
						}
					}
				}
			}break;
		}
		m_execWin->setSelectedExecGrille(m_currentGrille);
	}
}

void Vjpirate::clearInputs()
{

	m_inputs.clear();

	m_oscInputs.clear();
	m_midiCCInputs.clear();
	m_midiNoteInputs.clear();
	m_midiNoteRangeInputs.clear();
	for(unsigned int i=0;i<FEATURES_NB;++i) {
		m_audioInputs[i].clear();
	}

	for(unsigned int i=0;i<FEATURES_NB;++i) {
		m_loopAudioInputs[i].clear();
	}
}


void Vjpirate::triggerLoop(LOOP_ACTION action, int loop)
{
	timeval curTime;
	gettimeofday(&curTime,NULL);
	long longCurTime = (long)(curTime.tv_sec-((int)((float)curTime.tv_sec/1000000.0)*1000000))*1000 + (long)((float)curTime.tv_usec/1000.0);

	//if selected (currentGrille ) loop
	if(loop==SELECTED_LOOP) {
		Grille& currentGrille = m_grilles[m_currentGrille];
		currentGrille.triggerLoop(action, longCurTime, LOOP_TRIGGER_OVERDUB);
		int selLoop=SELECTED_LOOP;
		m_execWin->setLoopState(selLoop, currentGrille.getLoopState());
	} else { //standard loops
		if(loop<m_loops.size()) {
			//if this loop is empty: move or record
			if(m_loops[loop].getLoopState()==LOOP_EMPTY) {
				Grille& currentGrille = m_grilles[m_currentGrille];
				//if current Grille has a loop
				if(currentGrille.getLoopState()!=LOOP_EMPTY) {
					if(currentGrille.getLoopState()==LOOP_RECORDING || currentGrille.getLoopState()==LOOP_OVERDUBBING) {
						currentGrille.triggerLoop(LOOP_ACTION_STOP, longCurTime);
					}
					m_loops[loop].copy(currentGrille);
					m_loops[loop].setActive(true);
					//if audio activated on the loop
					if(currentGrille.getAudioActive()) {
						//test if one parameter has an audio input, add an audio input to the corresponding loop param
						for(unsigned int i=0;i<m_inputs.size();++i) {
							if(m_inputs[i].group==GRILLE_INPUT && (m_inputs[i].grille==m_currentGrille+1 || m_inputs[i].grille==SELECTED_GRILLE)) {
								if(m_inputs[i].type==INPUT_AUDIO) {
									Input newAudioInput = m_inputs[i];
									newAudioInput.group=LOOP_INPUT;
									newAudioInput.grille=loop;
									m_loopAudioInputs[m_inputs[i].audioFeature].push_back(newAudioInput);
								}
							}
						}
					}
					//erase loop of selected grille and desactive it		
					currentGrille.eraseLoop();
					currentGrille.setActive(false);
					int selGrille=SELECTED_GRILLE;
					setInputStoredValue(GRILLE_INPUT, selGrille, int(PARAMS_NB+PERMUT_PARAMS_NB+PARAM_ACTIVATION), 0);
					m_execWin->setCurrentGrilleAct(false);
					m_execWin->setLoopState(loop,m_loops[loop].getLoopState());
					int selLoop=SELECTED_LOOP;
					m_execWin->setLoopState(selLoop, currentGrille.getLoopState());
					setInputStoredValue(LOOP_INPUT, loop, int(LOOP_TRIGGER), 1);
				} else { 
					currentGrille.triggerLoop(action, longCurTime);
					m_execWin->setLoopState(loop,currentGrille.getLoopState());
				}
			}
			else { //if this loop is not empty, simply trigger it
				m_loops[loop].triggerLoop(action, longCurTime);
				m_execWin->setLoopState(loop, m_loops[loop].getLoopState());
			}
		}
	}
}

void Vjpirate::eraseLoop(int loop)
{
	if(loop==SELECTED_LOOP) {
		Grille& currentGrille = m_grilles[m_currentGrille];
		currentGrille.eraseLoop();
		m_execWin->setLoopState(loop, currentGrille.getLoopState());
		setInputStoredValue(APPLICATION_INPUT, -1, int(PARAM_LOOP_RECORD), 0);
	} else {
		if(m_loops[loop].getLoopState()==LOOP_EMPTY) {
			m_grilles[m_currentGrille].eraseLoop();
			m_grilles[m_currentGrille].setActive(false);
			int selGrille=SELECTED_GRILLE;
			setInputStoredValue(GRILLE_INPUT, selGrille, int(PARAM_LOOP_RECORD), 0);
		} else {
			m_loops[loop].eraseLoop();
			setInputStoredValue(LOOP_INPUT, loop, int(LOOP_TRIGGER), 0);
			m_loops[loop].setActive(false);
		}

		for(unsigned int f=0;f<FEATURES_NB;++f) {
			for(unsigned int i=0;i<m_loopAudioInputs[f].size();) {
				if(m_loopAudioInputs[f][i].group==LOOP_INPUT && m_loopAudioInputs[f][i].grille==loop) {
						m_loopAudioInputs[f].erase(m_loopAudioInputs[f].begin()+i);
				}
				else {
					++i;
				}
			}
		}
		m_execWin->setLoopState(loop,m_loops[loop].getLoopState());
	}
}

void Vjpirate::setLoopTransparency(int loop, float val)
{
	m_loops[loop].setLoopTransparency(val);
	m_execWin->setLoopTransparency(loop,val*100.0);
}

void Vjpirate::setLoopSpeed(int loop, float val, bool adjust)
{
	if(adjust){
		val=pow(2.0,2.0*val-1.0);
	}
	m_loops[loop].setLoopSpeed(val);
	m_execWin->setLoopSpeed(loop,val*100.0);
}


void Vjpirate::setMasterTransparency(float trans)
{
	//set transparency for all loops and grilles
	for(unsigned int loop=0;loop<m_loops.size();++loop) {
		m_loops[loop].setLoopTransparency(trans);
	}
	for(unsigned int gri=0;gri<m_grilles.size();++gri) {
		m_grilles[gri].setLoopTransparency(trans);
	}
	m_execWin->setMasterTransparency((trans-1.0)*-1.0);
}

void Vjpirate::setSelectActive(bool act)
{
	m_selectActive = act;
	m_execWin->setSelectActive(act);
}



