/***************************************************************************
 *            Vjpirate.hpp
 *
 *  2006  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifndef Vjpirate_h
#define Vjpirate_h

#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include <map>
#include <pthread.h>
#include <string>


#include "../input/InputManager.hpp"
#include "../ui/ExecWindow.hpp"
#include "../ui/RenderWindow.hpp"
#include "Grille.hpp"

enum VJ_APP_PARAM{PARAM_MASTER, PARAM_SELECT, PARAM_ACTIVE_SELECT, PARAM_LOOP_RECORD, PARAM_LOOP_ERASE};

class Vjpirate  
{	
	public:
		static Vjpirate* getInstance();
		~Vjpirate();
		void start();
		void stop();
        void update();

		void openFile(const char*, bool wait=true);
		void saveToFile(const char*);
		inline std::string getFileName(){return m_fileName;}
		inline std::string getShortFileName() {
			std::string shortFn=m_fileName;
			size_t pos = shortFn.find_last_of("/"); 
			if(pos!=shortFn.npos) {
				shortFn=shortFn.substr(pos+1);	
			}
			return shortFn;
		}
		void importInputs(const char*);

		void showMessage(std::string);

		Grille& addGrille();
		void removeGrille(int);
		void swapGrilles(int,int);
		Grille& duplicateGrille(int);
		void clearGrilles();
		inline Grille& getGrille(unsigned int gri){return m_grilles[gri];}
		inline unsigned int getNbGrilles(){return m_grilles.size();}
		void renameGrille(int,std::string);

		inline Grille& getCurrentGrille() { return m_grilles[m_currentGrille];}
		inline int getCurrentGrilleID() {return m_currentGrille;}
		
		void setCurrentGrille(unsigned int);

		void setCurrentGrilleAudioActive(bool);

		inline void setCurrentGrilleParameterValue(const VJ_PARAM& p, const float& v){m_grilles[m_currentGrille].setParameterValue(p,v);}
		inline void setCurrentGrilleParameterPatternPermutation(const VJ_PARAM& p, const float& v){m_grilles[m_currentGrille].setParameterPatternPermutation(p,v);}

		void loop();
		void clearLoops();
		inline Grille& getLoop(int loop){return m_loops[loop];}
		inline unsigned int getNbLoops(){return m_loops.size();}
		inline void setNbLoops(int nb){if(nb>0){m_loops.resize(nb,Grille());} }
		void resetGrillesGlCoords();

		//Anims
		Anim* openAnim(std::vector<std::string>);
		Anim* openAnim(std::string);
		void closeAnim(std::string);
		void testAllAnims();
		void clearAllAnims();
		inline unsigned int getNbAnims(){return m_anims.size();}
		inline Anim* getAnim(int a){return m_anims[a];}
		GLuint getNewTextureID();
		void freeTextureID(GLuint);

		//inputs
        void processAnalysis(double*);
		inline unsigned int getNbInputs(){return m_inputs.size();}
		inline Input& getInput(const int& inp){return m_inputs[inp];}
		void addInput(Input &inp);
		void addDefaultInput();
		void setInput(const int& input, Input &inp, const bool& learning=false);
		void removeInput(int);
		void learnInput(const bool&, const std::vector<int>&);
		void resetInputs(const INPUT_TYPE&, const bool& wait=true);
		void resetAllInputs();
		void clearInputs();
		void setInputStoredValue(const INPUT_GROUP&, const int&, const int&, const float&);
		void resetAudioInputStoredValues(const INPUT_GROUP&, const int&);
		void processInputEvent(const INPUT_TYPE&);

		inline InputEvent& getInputEvent(const INPUT_TYPE& type){return m_inputEvents[type];}

		bool isParameterofCurrentGrilleControlledByAudio(int);
		bool isParameterPermutofCurrentGrilleControlledByAudio(int);

		//loops
		friend void* loopThreadProcess(void*);
		void triggerLoop(LOOP_ACTION,int);
		void eraseLoop(int);
		void setLoopTransparency(int,float);
		void setLoopSpeed(int,float,bool adjust=true);

		void setMasterTransparency(float);
		void setSelectActive(bool);
		
		//redraw	
		friend void* redrawThreadProcess(void*);

		static const int SELECTED_GRILLE=0;
		static const int SELECTED_LOOP=-1;
		static const int LEARN_INPUT_CREATE=-1;

	private:

		Vjpirate();
		Vjpirate( const Vjpirate & ); 
    		Vjpirate & operator =( const Vjpirate & ); 

		//file
		std::string m_fileName;

		//ui
		ExecWindow *m_execWin;
		//redraw Thread
		pthread_t m_redrawThread;

		//grilles
		std::vector<Grille> m_grilles;
		int m_currentGrille;
		bool m_selectActive;

		//loops
		std::vector<Grille> m_loops;
		static const int NB_LOOPS=8;
		bool m_loopEraseToggle;

		//Loop Thread
		pthread_t m_loopThread;

		//controls
		InputManager* m_inputManager;
		std::vector<Input> m_inputs;
		std::map<std::string, std::vector<int> > m_oscInputs;
		std::map<int, std::vector<int> > m_midiCCInputs;
		std::map<char, std::vector<int> > m_keyboardInputs;
		std::map<int, std::vector<int> > m_midiNoteInputs;
		std::map<int, std::vector<int> > m_midiNoteRangeInputs;
		std::vector<int> m_audioInputs[FEATURES_NB];		

		bool m_resetInputsAsked[INPUT_TYPES_NB];

		bool m_learnInput;
		std::vector<int> m_learnInputArgs;

		std::vector<Input> m_loopAudioInputs[FEATURES_NB];
	
		InputEvent m_inputEvents[INPUT_TYPES_NB];

		//anims
		std::vector<Anim*> m_anims;

		std::vector<GLuint> m_textureIDs;
		std::vector<bool> m_textureIDUsed;
};

#endif

