/***************************************************************************
 *  InputManager.cpp
 *  Part of VjPirate
 *  2006  Tardigrade Inc.
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "InputManager.hpp"
#include "../base/Vjpirate.hpp"
#include <xtract/libxtract.h>
#include <jack/midiport.h>

using namespace osc;
using namespace std;

const char InputManager::MIDI_NOTE_ON=0x90;
const char InputManager::MIDI_NOTE_OFF=0x80;
const char InputManager::MIDI_PROGRAM_CHANGE=0xC0;
const char InputManager::MIDI_CONTROLLER=0xB0;

InputManager* InputManager::getInstance() {
	static InputManager instance;
	return &instance;
}

int jackCallback(jack_nframes_t nbFrames, void *arg) {
    InputManager* man = InputManager::getInstance();
    man->processJack(nbFrames);
	return 0;
}

InputManager::InputManager():m_defaultPort(8327) {
    //create the jack client
	m_jackClient = jack_client_open("VjPirate", JackUseExactName, NULL);	
    if(!m_jackClient) {
        cout<<"Error creating the jack client"<<endl;
    }
	m_sampleRate = jack_get_sample_rate(m_jackClient);	
	m_bufferSize = jack_get_buffer_size(m_jackClient);	

    //initialize the jack ringbuffer
    InputEvent ev;
    m_inputEventSize=sizeof(ev);
    m_audioRingbuffer = jack_ringbuffer_create(1000*m_inputEventSize);
    jack_ringbuffer_mlock(m_audioRingbuffer);
    m_midiRingbuffer = jack_ringbuffer_create(1000*m_inputEventSize);
    jack_ringbuffer_mlock(m_midiRingbuffer);

    //initialize audio analysis
    m_analysisSize=1024;
    m_meanSize=10;
    xtract_init_fft(m_analysisSize, XTRACT_SPECTRUM);
    m_barkBands = new int[XTRACT_BARK_BANDS];
    xtract_init_bark(m_bufferSize, m_sampleRate, m_barkBands);
    m_inputBuffer.resize(m_bufferSize,0);
    m_spectrumBuffer.resize(m_bufferSize*2.0,0);
    m_prevSpectrumBuffer.resize(m_bufferSize*2.0,0);
    m_barkBuffer.resize(XTRACT_BARK_BANDS,0);
    deque<double> feat;
    m_featuresAvg.resize(FEATURES_NB,feat);
    m_featuresSums.resize(FEATURES_NB,0);
    for(unsigned int i=0;i<FEATURES_NB;++i) {
        m_featuresAvg[i].resize(int(m_meanSize),0);
    }

    //create the audio and midi ports
    m_audioPort=jack_port_register(m_jackClient, "AudioInput",
                                            JACK_DEFAULT_AUDIO_TYPE, 
                                            JackPortIsInput, 0);
    m_midiPort=jack_port_register(m_jackClient, "MidiInput",
                                            JACK_DEFAULT_MIDI_TYPE,
                                            JackPortIsInput, 0);

	//process callback
	jack_set_process_callback (m_jackClient, jackCallback, this);

	//activate the client
	jack_activate(m_jackClient);

    //create the osc socket 
    try {
        m_socket = new UdpListeningReceiveSocket(
                            IpEndpointName(IpEndpointName::ANY_ADDRESS, 8327),
                            this);
        m_address = new char[IpEndpointName::ADDRESS_STRING_LENGTH];
        MAX_BUFFER_SIZE=4096;
        m_data = new char[MAX_BUFFER_SIZE];
    }
    catch(const exception& e) {
        DEBUG("Error creating the osc input "<<e.what());
    }
    cout<<"Input Manager created"<<endl;
}

InputManager::~InputManager() {
    jack_deactivate(m_jackClient);
    jack_client_close(m_jackClient);
}

void InputManager::processJack(const jack_nframes_t& nbFrames) {
    //audio port
    jack_default_audio_sample_t* buf
        = (jack_default_audio_sample_t *) 
            jack_port_get_buffer(m_audioPort, nbFrames);
    InputEvent newAudioEvent;
    for(jack_nframes_t f=0; f<nbFrames && f<InputEvent::maxBufferSize; ++f) {
        newAudioEvent.buffer[f]=buf[f];
    }
    if(jack_ringbuffer_write_space(m_audioRingbuffer)>m_inputEventSize) {
        jack_ringbuffer_write(m_audioRingbuffer, 
                                (char* )&newAudioEvent, m_inputEventSize);
    }

    //midi port
    void* buffer = jack_port_get_buffer(m_midiPort, nbFrames);
    unsigned int nbEvents = jack_midi_get_event_count(buffer);
    jack_midi_event_t event;
    for(unsigned int e=0; e<nbEvents; ++e) {
        if(!jack_midi_event_get(&event, buffer, e)) {
            if(event.size==3) {
                InputEvent newMidiEvent;
                char message = (event.buffer[0]>>4)<<4;
                newMidiEvent.midiChannel=(char)(message^event.buffer[0]);
                newMidiEvent.midiParam=(char)(event.buffer[1]);
                newMidiEvent.value=float(event.buffer[2])/127.0;
                if(message==MIDI_NOTE_OFF 
                        || (message==MIDI_NOTE_ON 
                                && newMidiEvent.value==0)) {
                    newMidiEvent.value=0;
                    newMidiEvent.type=INPUT_MIDI_NOTE;
                    if(jack_ringbuffer_write_space(m_midiRingbuffer)
                                                    >m_inputEventSize) {
                        jack_ringbuffer_write(m_midiRingbuffer, 
                                                (char* )&newMidiEvent, 
                                                m_inputEventSize);
                    }
                }
				else if(message == MIDI_NOTE_ON) {
                    newMidiEvent.type=INPUT_MIDI_NOTE;
                    if(jack_ringbuffer_write_space(m_midiRingbuffer)
                                                    >m_inputEventSize) {
                        jack_ringbuffer_write(m_midiRingbuffer, 
                                                (char* )&newMidiEvent, 
                                                m_inputEventSize);
                    }
                    newMidiEvent.type=INPUT_MIDI_NOTE_RANGE;
                    newMidiEvent.value=float(event.buffer[2]);
                    if(jack_ringbuffer_write_space(m_midiRingbuffer)
                                                    >m_inputEventSize) {
                        jack_ringbuffer_write(m_midiRingbuffer, 
                                                (char* )&newMidiEvent, 
                                                m_inputEventSize);
                    }
                }
				else if(message == MIDI_CONTROLLER) {
                    newMidiEvent.type=INPUT_MIDI_CC;
                    if(jack_ringbuffer_write_space(m_midiRingbuffer)
                                                    >m_inputEventSize) {
                        jack_ringbuffer_write(m_midiRingbuffer, 
                                                (char* )&newMidiEvent, 
                                                m_inputEventSize);
                    }
                }
            }
        }
    }
}

void InputManager::ProcessMessage( const osc::ReceivedMessage& m, 
                                    const IpEndpointName& remoteEndpoint ) {
	if(m.ArgumentCount()>0) { 
		ReceivedMessage::const_iterator arg = m.ArgumentsBegin();
		std::string mess(m.AddressPattern());
		if(arg->IsFloat() || arg->IsInt32()) {
			float val=0;
			if(arg->IsFloat()) {
				val = (float)((arg++)->AsFloatUnchecked());
			}
			else if( arg->IsInt32() ) {
				val = (float)((arg++)->AsInt32Unchecked());
			}

			Vjpirate *vjp = Vjpirate::getInstance();
			InputEvent& event = vjp->getInputEvent(INPUT_OSC);
			event.oscMessage=mess;
			event.value=val;
			vjp->processInputEvent(INPUT_OSC);
		}
	}
}

void InputManager::start() {

}

void InputManager::stop() {
}

void InputManager::process() {
	Vjpirate *vjp = Vjpirate::getInstance();

/*
    //process osc 
    std::size_t size = m_socket->ReceiveFrom(m_remoteEndPoint, m_data, 
                                                MAX_BUFFER_SIZE);
    if(size>0) {
        ProcessPacket(m_data, (int)size, m_remoteEndPoint);
    }
    cout<<"inputmanager processed osc"<<endl;
*/
    //process audio
    InputEvent newAudioEvent;
    while(jack_ringbuffer_read_space(m_audioRingbuffer)>=m_inputEventSize) {
        jack_ringbuffer_read(m_audioRingbuffer,
                                (char*)&newAudioEvent, m_inputEventSize);
        for(unsigned int b=0; b<m_inputBuffer.size(); ++b) {
            m_inputBuffer[b]=newAudioEvent.buffer[b];
        }
    }

    float argf[4];
    argf[0] = 8000.f;
    argf[1] = XTRACT_MAGNITUDE_SPECTRUM;
    argf[2] = 0.f;
    argf[3] = 0.f;

    xtract_spectrum(&(m_inputBuffer[0]), m_inputBuffer.size(), &argf[0], &(m_spectrumBuffer[0]));
    xtract_bark_coefficients(&(m_spectrumBuffer[0]), m_spectrumBuffer.size()/2, m_barkBands, &(m_barkBuffer[0]));
    m_features[ANALYSIS_BASS]=0;
    m_features[ANALYSIS_MEDIUM]=0;
    m_features[ANALYSIS_HIGH]=0;
    for(unsigned int i=0;i<XTRACT_BARK_BANDS/3;++i) {
        m_features[ANALYSIS_BASS] += m_barkBuffer[i];
        m_features[ANALYSIS_MEDIUM] += m_barkBuffer[XTRACT_BARK_BANDS/3+i];
        m_features[ANALYSIS_HIGH] += m_barkBuffer[2*XTRACT_BARK_BANDS/3+i];
    }

    m_features[ANALYSIS_BASS]*=3;
    m_features[ANALYSIS_MEDIUM]*=3;
    m_features[ANALYSIS_HIGH]*=3;


    xtract_rms_amplitude(&(m_inputBuffer[0]), m_inputBuffer.size(),
                            NULL, &m_features[ANALYSIS_RMS]);

    //extract the other features only if there is enough loudness
    if(m_features[ANALYSIS_RMS]>0.05) {
        //cout<<"rms "<<features[ANALYSIS_RMS]<<endl;
        xtract_flatness(&(m_spectrumBuffer[0]),
                            m_spectrumBuffer.size()/2, NULL,
                            &m_features[ANALYSIS_FLATNESS]);
        xtract_spectral_centroid(&(m_spectrumBuffer[0]),
                                m_spectrumBuffer.size()/2, NULL,
                                &m_features[ANALYSIS_CENTROID]);
        m_features[ANALYSIS_CENTROID] /= 100000.0;
        m_features[ANALYSIS_FLATNESS] *= 1000.0;
    }

    //low-pass filter the features, except the onset
    for(unsigned int i=0;i<FEATURES_NB-1;++i) {
        m_featuresSums[i]-=m_featuresAvg[i][0];
        m_featuresAvg[i].pop_front();
        m_featuresSums[i]+=m_features[i];
        m_featuresAvg[i].push_back(m_features[i]);
        m_features[i]=m_featuresSums[i]/m_meanSize;
    }

    extractOnset(m_spectrumBuffer, m_prevSpectrumBuffer, 
                    m_features[ANALYSIS_ONSET]);
    m_prevSpectrumBuffer=m_spectrumBuffer;

    vjp->processAnalysis(m_features);


    //MIDI events
    InputEvent newMidiEvent;
    while(jack_ringbuffer_read_space(m_midiRingbuffer)>=m_inputEventSize) {
        jack_ringbuffer_read(m_midiRingbuffer,
                                (char*)&newMidiEvent, m_inputEventSize);
        InputEvent& event = vjp->getInputEvent(newMidiEvent.type);
        event=newMidiEvent;
        vjp->processInputEvent(newMidiEvent.type);
    }
}

void InputManager::extractOnset(std::vector<double>& spectrumBuffer, 
                               std::vector<double>& prevSpectrumBuffer, 
                                double& onset) {
    //FIXME add the peak detection step
    onset=0;
    for(unsigned int f=0; f<spectrumBuffer.size()/2; ++f) {
         onset+=0.5*(fabs(spectrumBuffer[f])-fabs(prevSpectrumBuffer[f])
                 +fabs(fabs(spectrumBuffer[f])-fabs(prevSpectrumBuffer[f])));
    }
    if(onset>=0.1) {
        onset=1;
    }
    else {
        onset=0;
    }
}


