/***************************************************************************
 *  InputManager.hpp
 *  Part of VjPirate
 *  2006  Tardigrade Inc.
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef INPUT_MANAGER_H
#define INPUT_MANAGER_H
#include <iostream>
#include <math.h>
#include <vector>
#include <deque>
#include <stdlib.h>
#include "oscPack/ip/UdpSocket.h"
#include "oscPack/ip/IpEndpointName.h"
#include "oscPack/osc/OscReceivedElements.h"
#include "oscPack/osc/OscPacketListener.h"
#include <jack/jack.h>
#include <jack/types.h>
#include <jack/ringbuffer.h>

enum INPUT_TYPE {INPUT_KEYBOARD, INPUT_MIDI_CC, INPUT_MIDI_NOTE, INPUT_MIDI_NOTE_RANGE, INPUT_OSC, INPUT_AUDIO, INPUT_TYPES_NB};
enum INPUT_MODE {MODE_NORMAL, MODE_TOGGLE, MODE_STEP_NORMAL, MODE_STEP_CIRCULAR, MODE_STEP_PING, MODE_STEP_RANDOM, INPUT_MODES_NB};
enum INPUT_GROUP {GRILLE_INPUT, LOOP_INPUT, APPLICATION_INPUT};
enum ANALYSIS_FEATURE {ANALYSIS_NONE=-1, ANALYSIS_RMS=0, ANALYSIS_BASS=1, ANALYSIS_MEDIUM=2, ANALYSIS_HIGH=3, ANALYSIS_FLATNESS=4, ANALYSIS_CENTROID=5, ANALYSIS_ONSET=6, FEATURES_NB=7};

struct Input {
	Input(){prevOut=0;}
	int group;
	int grille;
	int parameter;
	INPUT_TYPE type;
	int midiChannel;
	int midiCC;
	int midiNote;
	int midiNoteMax;
    ANALYSIS_FEATURE audioFeature;
	char key;
	std::string oscMessage;	
	INPUT_MODE mode;
	float step;

	float prevOut;

	void processValue(const float & in, float & out) {
		switch(mode) {
			case MODE_NORMAL: { out=in;}break;
			case MODE_TOGGLE: { 
				if(in>0) {
					out=(prevOut>0)?0:1;
					prevOut=out;
				}
				else {
					out=-1;	
				}
			}break;
			case MODE_STEP_NORMAL: {
				out=prevOut;
				if(in>0) {
					out=prevOut+step;
					if(out>1) {
						out=1;
					}
					else if(out<0) {
						out=0;
					}
					prevOut=out;
				}
			}break;
			case MODE_STEP_CIRCULAR: {
				out=prevOut;
				if(in>0) {
					out=prevOut+step;
					if(out>1) {
						out=0;
					}
					else if(out<0) {
						out=1;
					}
					prevOut=out;
				}
			}break;
			case MODE_STEP_PING: {
				out=prevOut;
				if(in>0) {
					out=prevOut+step;
					if(out>1) {
						out=1;
						step=-step;
					}
					else if(out<0) {
						out=0;
						step=-step;
					}
					prevOut=out;
				}
			}break;
			case MODE_STEP_RANDOM: {
				out=prevOut;
				if(in>0) {
					out = 1.0/float((rand()%int(1.0/step))+1);
					prevOut=out;
				}
			}break;
			default:{
			}break;
		}
	}
};

struct InputEvent {
	INPUT_TYPE type;
	std::string oscMessage;
	int midiChannel;
	int midiParam;
	char key;
	float value;
    ANALYSIS_FEATURE audioFeature;
    static const int maxBufferSize=1024;
    float buffer[maxBufferSize];
};


class InputManager : public osc::OscPacketListener {
	public:
		static InputManager* getInstance();
		~InputManager();
		void start();
		void stop();
        void process();

		void ProcessMessage(const osc::ReceivedMessage&, const IpEndpointName&);
        void processJack(const jack_nframes_t& nbFrames);

    private:
		friend int jackCallback(jack_nframes_t , void *);
        void extractOnset(std::vector<double>& spectrumBuffer, 
                           std::vector<double>& prevSpectrumBuffer, 
                            double& onset);
    
	private:
		InputManager();

        UdpListeningReceiveSocket* m_socket;
        int m_defaultPort;
        char* m_address;
        char* m_data;
        IpEndpointName m_remoteEndPoint;
        static const int IP_MTU_SIZE=1536;
        std::size_t MAX_BUFFER_SIZE;

		jack_client_t *m_jackClient;
		unsigned int m_sampleRate;	
		unsigned int m_bufferSize;	
        jack_port_t* m_audioPort;
        jack_port_t* m_midiPort;
        jack_ringbuffer_t*  m_audioRingbuffer;
        jack_ringbuffer_t*  m_midiRingbuffer;
        size_t m_inputEventSize;
        int m_analysisSize;
        double m_features[FEATURES_NB];
        float m_meanSize;
        int* m_barkBands;
        std::vector<std::deque<double> >  m_featuresAvg;
        std::vector<double> m_featuresSums;
        std::vector<double> m_inputBuffer;
        std::vector<double> m_spectrumBuffer;
        std::vector<double> m_prevSpectrumBuffer;
        std::vector<double> m_barkBuffer;

        static const char MIDI_NOTE_ON;
        static const char MIDI_NOTE_OFF;
        static const char MIDI_PROGRAM_CHANGE;
        static const char MIDI_CONTROLLER;


};






#endif 
